import cobra
import pandas as pd
from cobra.io import load_json_model
from glob import glob
from cobra import Reaction, Metabolite
from Bio import SeqIO

# load the base model
master_model = load_json_model('model_generation/iYL1228_formulae_core.json')
# remove all the GPRs for each reaction, but save them in a dictionary so we can add them back
# to the master model with the GPRs later
master_gprs = {}
for reaction in master_model.reactions:
    master_gprs[reaction.id] = reaction.gene_reaction_rule
    reaction.gene_reaction_rule = ''

# set up list of new reactions to add
new_reactions = []
# set up a list of new metabolites to add
new_metabolites = {}
# set up dict of new reactions and their gprs
new_reactions_gprs = {}

# read in the spreadsheet with the reactions and metabolites
rxns = pd.read_excel('master_model_reaction_table.xlsx', sheet_name='Reactions')
metabs = pd.read_excel('master_model_reaction_table.xlsx', sheet_name='Metabolites')

# create a list of metabolites first so we can add the reactions
for index,row in metabs.iterrows():
    new_metab = Metabolite(row['metabolite'], name=row['name'], formula=row['formula'], compartment=row['compartment'])
    new_metabolites[row['metabolite']] = new_metab

# loop through each reaction and add it to the model
for index,row in rxns.iterrows():
    new_rxn = Reaction(row['reaction'])
    new_rxn.name = row['name']
    new_rxn.lower_bound = row['lower']
    new_rxn.upper_bound = row['upper']
    new_rxn.add_metabolites(eval(row['metabolite_dict']))
    new_reactions.append(new_rxn)
    # add the gpr to the dictionary
    if pd.isna(row['gpr']):
        new_reactions_gprs[row['reaction']] = ''
    else:
        new_reactions_gprs[row['reaction']] = row['gpr']

# add all the reactions in
master_model.add_reactions(new_reactions)

# update the reaction bounds for the following reactions, as they were incorrect in the original model
# update the tartrate, succinate, formic acid, fumaric acid, L-threonine,
# N-Acetyl-beta-D Mannosamine, D-Malic acid, alpha-Keto-Glutaric Acid and 4-Aminobutanoate transport reaction bounds
master_model.reactions.TARTRtex.upper_bound = 1000
master_model.reactions.SUCCtex.upper_bound = 1000
master_model.reactions.FORtex.upper_bound = 1000
master_model.reactions.FUMtex.upper_bound = 1000
master_model.reactions.THRtex.upper_bound = 1000
master_model.reactions.ACMANAtex.upper_bound = 1000
master_model.reactions.MALDtex.upper_bound = 1000
master_model.reactions.ABUTtex.upper_bound=1000
master_model.reactions.AKGtex.upper_bound = 1000
master_model.reactions.AKGtex.lower_bound = -1000

# write out
cobra.io.json.save_json_model(master_model,'Klebsiella_master_model.json', pretty=False)

# now we want to create a master version of the model that carries all the GPRs for each reaction
# this will include reactions where the gene homology is a wide distribution and so will include genes from multiple species

for reaction in master_model.reactions:
    if reaction.id in master_gprs.keys():
        # add the gene reaction rule back from the original model
        reaction.gene_reaction_rule = master_gprs[reaction.id]
    # if the reaction wasn't in the original model, add it from the table
    elif reaction.id in new_reactions_gprs.keys():
        reaction.gene_reaction_rule = new_reactions_gprs[reaction.id]
    else:
        print(reaction.id, ' has no gpr listed anywhere')

# update the GPRs for reactions where the gene homology drops below 75%, but we have
# variant genes in the other models
reaction_gpr_dict = {'DDGALK': 'KPN_04097 or OAIMIDIL_01955', 'GALCTNt2pp': 'KPN_04094 or HMPHIOOD_04301', 'HPACt2r': 'KPN_04782 or HMPHIOOD_04460', 'MELIBt2pp': 'KPN_04506 or EHJJNFNP_00706', 'ARBt2rpp': 'KPN_03264 or AFBMIAHL_04773'}
for reaction in master_model.reactions:
    if reaction.id in reaction_gpr_dict.keys():
        reaction.gene_reaction_rule = reaction_gpr_dict[reaction.id]

# create a new biomass function that removes the udpgalur_c and udpgal_c metabolites
new_biomass = master_model.reactions.get_by_id('BIOMASS_Core')
new_biomass.name = 'BIOMASS_Core_Oct2019'
new_biomass.id = 'BIOMASS_Core_Oct2019'

master_model.add_reaction(new_biomass)

master_model.reactions.get_by_id('BIOMASS_Core_Oct2019').subtract_metabolites({master_model.metabolites.get_by_id('udpgalur_c'): -0.001})
master_model.reactions.get_by_id('BIOMASS_Core_Oct2019').subtract_metabolites({master_model.metabolites.get_by_id('udpgal_c'): -0.076})

# make sure we set the objective to be the new biomass function
master_model.objective = 'BIOMASS_Core_Oct2019'

# save the model
cobra.io.json.save_json_model(master_model,'Klebsiella_master_model_GPRs.json', pretty=False)

# now create a list of genes - we need to go into the fasta files and extract the genes we want
gene_list = []
for gene in master_model.genes:
    gene_list.append(gene.id)

# generate a list of protein sequences for all the genes that are in our GPR list
prot_seqs = []
prot_files = glob('models_38Genomes/models_141019/prot/*.fa')
for prot in prot_files:
    prot_in = SeqIO.parse(prot, 'fasta')
    for record in prot_in:
        if record.id in gene_list:
            prot_seqs.append(record)

# do the same for the nucleotide sequences
nucl_seqs = []
nucl_files = glob('models_38Genomes/models_141019/nucl/*.fa')
for nucl in nucl_files:
    nucl_in = SeqIO.parse(nucl, 'fasta')
    for record in nucl_in:
        if record.id in gene_list:
            nucl_seqs.append(record)

# write out those sequences
# NOTE there is one fewer gene because there is no KPN_SPONT gene, as this is a fake gene for spontaneous reactions
SeqIO.write(prot_seqs, 'Klebsiella_master_model_proteinSeqs.fa', 'fasta')
SeqIO.write(nucl_seqs, 'Klebsiella_master_model_nuclSeqs.fa', 'fasta')
